/* The Monitor
 * Copyright (C) 1998 Oskar Liljeblad
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "gtk_ui.h"
#include "monitor.h"
#include "util.h"
#include "ui.h"

static void destroy_event (GtkWidget *, gpointer);
static gint delete_event (GtkWidget *, GdkEvent *, gpointer);
static void input_event (gpointer, gint, GdkInputCondition);
static gint input_timer (gpointer);
static GdkColor *allocate_color (char *);

static GtkWidget *text;
static GtkWidget *window;

struct arg_options ui_opts[] = {
/* per-watch userinterface options */
	{ "foreground", 'f', 	ARG_REQUIRED, FALSE, NULL,		"=COLOR", "text foreground color (see X(1) for format)" },
	{ "no-foreground", 0, 	ARG_NONE, 	  FALSE, NULL, 		NULL, "revert to default foreground color" },
	{ "background", 'b', 	ARG_REQUIRED, FALSE, NULL,		"=COLOR", "text background color (see X(1) for format)" },
	{ "no-background", 0, 	ARG_NONE, 	  FALSE, NULL, 		NULL, "revert to default background color" },
	{ "font", 0, 			ARG_REQUIRED, FALSE, NULL, 		"=FONT", "text font" },
	{ "no-font", 0, 		ARG_NONE, 	  FALSE, NULL, 		NULL, "revert to default font" },
/* userinterface options */
	{ "geometry", 0, 		ARG_REQUIRED, FALSE, NULL, 		NULL, "window size and position (see X(1) for format)" },
	{ "hscrollbar", 0, 		ARG_NONE, 	  FALSE, NULL, 		NULL, "display horizontal scrollbar" },
	{ "no-hscrollbar", 0, 	ARG_NONE, 	  FALSE, NULL, 		NULL, "do not display horizontal scrollbar" },
	{ "vscrollbar", 0, 		ARG_NONE, 	  TRUE,  NULL, 		NULL, "display vertical scrollbar" },
	{ "no-vscrollbar", 0, 	ARG_NONE, 	  FALSE, NULL, 		NULL, "do not display vertical scrollbar" },
	{ NULL, 0, 0, 0, NULL, NULL, NULL }
};

char *ui_base_name = "gtkmon";

int
initialize_ui (argc, argv)
  int *argc;
  char ***argv;
{
	gtk_init(argc, argv);

	return TRUE;
}

int
enter_ui_mainloop (opts)
  struct arg_options *opts;
{
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *scrollbar_v;
	GtkWidget *scrollbar_h;

	/* create window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), PACKAGE);
	gtk_container_border_width(GTK_CONTAINER(window), 0);
	gtk_window_set_policy (GTK_WINDOW(window), TRUE, TRUE, FALSE);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(destroy_event), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(delete_event), NULL);
	// circles = gdk_bitmap_create_from_data (window->window, circles_bits, circles_width, circles_height);
	// gdk_window_set_icon (window->window, NULL, circles, circles);
	gtk_widget_realize(window);
	gdk_window_set_icon_name (window->window, PACKAGE);

	/* create containers */
	vbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_widget_show(vbox);
	hbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	gtk_widget_show(hbox);

	/* create text widget */
	text = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_text_set_word_wrap(GTK_TEXT(text), TRUE);
	GTK_WIDGET_UNSET_FLAGS(text, GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 0);
	gtk_widget_realize(text);
	gtk_text_set_point(GTK_TEXT(text), 0);
	gtk_widget_show(text);

	/* create horizontal scrollbar */
	if (opts[OPT_HSCROLLBAR].exists) {
		scrollbar_h = gtk_hscrollbar_new(GTK_TEXT(text)->hadj);
		GTK_WIDGET_UNSET_FLAGS(scrollbar_h, GTK_CAN_FOCUS);
		gtk_box_pack_start(GTK_BOX(hbox), scrollbar_h, FALSE, FALSE, 0);
		gtk_widget_show(scrollbar_h);
	}

	/* create vertical scrollbar */
	if (opts[OPT_VSCROLLBAR].exists) {
		scrollbar_v = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
		gtk_box_pack_start(GTK_BOX(vbox), scrollbar_v, FALSE, FALSE, 0);
		GTK_WIDGET_UNSET_FLAGS(scrollbar_v, GTK_CAN_FOCUS);
		gtk_widget_show(scrollbar_v);
	}

	if (opts[OPT_GEOMETRY].exists) {
		XSizeHints *sizehints;
		GtkAllocation alloc;
		int res, x, y;
		unsigned int width, height;
		guint rootwidth, rootheight;

		gdk_window_get_size(GDK_ROOT_PARENT(), &rootwidth, &rootheight);
		if (!rootwidth || !rootheight) {
			printerror("Could not get root window geometry", NULL);
			return FALSE;
		}

		res = XParseGeometry(opts[OPT_GEOMETRY].string, &x, &y, &width, &height);
		if (!(res & WidthValue)) width = DEFAULT_WINDOW_WIDTH;
		if (!(res & HeightValue)) height = DEFAULT_WINDOW_HEIGHT;

		sizehints = XAllocSizeHints();
		if (!sizehints) {
			printerror("Memory exhausted", NULL);
			return FALSE;
		}

		if (res & XNegative && res & YNegative) {
			sizehints->win_gravity = SouthEastGravity;
			x = rootwidth - width;
			y = rootheight - height;
		} else if (res & XNegative) {
			sizehints->win_gravity = NorthEastGravity;
			x = rootwidth - width;
		} else if (res & YNegative) {
			sizehints->win_gravity = SouthWestGravity;
			y = rootheight - height;
		} else {
			sizehints->win_gravity = NorthWestGravity;
		}
		sizehints->flags = USPosition | PWinGravity;
		sizehints->x = x;
		sizehints->y = y;
		sizehints->width = width;
		sizehints->height = height;

		alloc.x = x;
		alloc.y = y;
		alloc.width = width;
		alloc.height = height;
		gtk_widget_size_allocate(window, &alloc);
		gtk_widget_set_uposition(window, x, y);
		gtk_widget_set_usize(window, width, height);

		gtk_widget_show(window);

		/* I don't want to write an error handler *just* to make sure
		 * this function is gracefully handled. GTK should take care of
		 * it, I hope.
		 */
		XSetWMNormalHints(GDK_WINDOW_XDISPLAY(window->window), GDK_WINDOW_XWINDOW(window->window), sizehints);
		XFree(sizehints);
	} else
		gtk_widget_show(window);

	gtk_main();

	return TRUE;
}

void
exit_ui (void)
{
	gtk_main_quit();
}

static gint
delete_event (widget, event, data)
  GtkWidget *widget;
  GdkEvent *event;
  gpointer data;
{
	return FALSE;
}

static void
destroy_event (widget, data)
  GtkWidget *widget;
  gpointer data;
{
	gtk_main_quit();
}

static void
input_event (data, source, condition)
  gpointer data;
  gint source;
  GdkInputCondition condition;
{
	if (condition == GDK_INPUT_READ)
		if (!update_watch(WATCH(data)))
			exit_ui();
}

static gint
input_timer (data)
  gpointer data;
{
	if (!update_watch(WATCH(data)))
		exit_ui();

	return TRUE;
}

int
monitor_file_event (watch)
  Watch *watch;
{
	UIWATCH(watch)->event_id = gdk_input_add(watch->fd, GDK_INPUT_READ, input_event, watch);
	/* if there's an error, gtk will usually take care of it */
	if (!UIWATCH(watch)->event_id)
		return FALSE;

	return TRUE;
}

void
unmonitor_file_event (watch)
  Watch *watch;
{
	gdk_input_remove(UIWATCH(watch)->event_id);
}

int
monitor_timer (watch)
  Watch *watch;
{
	UIWATCH(watch)->event_id = gtk_timeout_add(watch->delay, input_timer, watch);
	/* if there's an error, gtk will usually take care of it */
	if (!UIWATCH(watch)->event_id)
		return FALSE;
	
	return TRUE;
}

void
unmonitor_timer (watch)
  Watch *watch;
{
	gtk_timeout_remove(UIWATCH(watch)->event_id);
}

static GdkColor *
allocate_color (str)
  char *str;
{
	GdkColormap *colormap;
	GdkColor *color;
	XColor xcolor;

	colormap = gdk_colormap_get_system();
	/* colormap = gtk_widget_get_colormap(text); */

	color = malloc(sizeof(GdkColor));
	if (!color) {
		printerror("Memory exhausted", NULL);
		return NULL;
	}

	/* GTK should take care of this function if it fails */
	if (!XParseColor (GDK_COLORMAP_XDISPLAY(colormap), GDK_COLORMAP_XCOLORMAP(colormap), str, &xcolor)) {
		printerror("could not look up color `", str, "'", NULL);
		free(color);
		return NULL;
	}

	color->pixel = xcolor.pixel;
	color->red = xcolor.red;
	color->green = xcolor.green;
	color->blue = xcolor.blue;

	/* GTK should print error if necessary, I hope */
	if (!gdk_color_alloc(colormap, color)) {
		printerror("could not allocate color `", str, "'", NULL);
		free(color);
		return NULL;
	}

	return color;
}

int
ui_parse_args_callback(str, opt, opts)
  char *str;
  int opt;
  struct arg_options *opts;
{
	static char *timestamp_fg = NULL;
	static char *timestamp_bg = NULL;
	static char *timestamp_font = NULL;

	if (opts) {
		if (str) {
			UIWatch *watch;
			watch = (UIWatch *) new_watch(str, sizeof(UIWatch)-sizeof(Watch));
			if (!watch)
				return FALSE;
			/* set text details */
			if (opts[OPT_FOREGROUND].string && !(watch->fg = allocate_color(opts[OPT_FOREGROUND].string))) {
				return FALSE;
			}
			if (opts[OPT_BACKGROUND].string && !(watch->bg = allocate_color(opts[OPT_BACKGROUND].string))) {
				return FALSE;
			}
			if (opts[OPT_FONT].string && !(watch->font = gdk_font_load(opts[OPT_FONT].string))) {
				printerror("could not load font `", opts[OPT_FONT].string, "'", NULL);
				return FALSE;
			}

			/* set timestamp details */
			if (timestamp_fg && !(watch->timestamp_fg = allocate_color(timestamp_fg))) {
				return FALSE;
			}
			if (timestamp_bg && !(watch->timestamp_bg = allocate_color(timestamp_bg))) {
				return FALSE;
			}
			if (timestamp_font && !(watch->timestamp_font = gdk_font_load(timestamp_font))) {
				printerror("could not load font `", timestamp_font, "'", NULL);
				return FALSE;
			}
		} else {
			if (opt == OPT_TIMESTAMP) {
				timestamp_fg = opts[OPT_FOREGROUND].string;
				timestamp_bg = opts[OPT_BACKGROUND].string;
				timestamp_font = opts[OPT_FONT].string;
			} else if (opt == OPT_NO_FOREGROUND
			  || opt == OPT_NO_BACKGROUND
			  || opt == OPT_NO_FONT
			  || opt == OPT_NO_HSCROLLBAR
			  || opt == OPT_NO_VSCROLLBAR) {
				opts[opt-1].string = NULL;
				opts[opt-1].exists = FALSE;
			}
		}
	}

	return TRUE;
}

int
ui_print_line (watch, str, len, type)
  Watch *watch;
  char *str;
  unsigned int len;
  int type;
{
	if (type == WATCH_FLAG_TIMESTAMP)
		gtk_text_insert(GTK_TEXT(text), UIWATCH(watch)->timestamp_font, UIWATCH(watch)->timestamp_fg, UIWATCH(watch)->timestamp_bg, str, len);
	else
		gtk_text_insert(GTK_TEXT(text), UIWATCH(watch)->font, UIWATCH(watch)->fg, UIWATCH(watch)->bg, str, len);

	return TRUE;
}

void ui_close_watch(watch)
  Watch *watch;
{
	/* not necessary to unload fonts - watch->font, watch->timestamp_font (?) */
	if (UIWATCH(watch)->fg) free(UIWATCH(watch)->fg);
	if (UIWATCH(watch)->bg) free(UIWATCH(watch)->bg);
	if (UIWATCH(watch)->timestamp_fg) free(UIWATCH(watch)->timestamp_fg);
	if (UIWATCH(watch)->timestamp_bg) free(UIWATCH(watch)->timestamp_bg);
}
