/* The Monitor
 * Copyright (C) 1998 Oskar Liljeblad
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "monitor.h"
#include "util.h"
#include "ui.h"
#include "common.h"

inline static int parse_args(int, char **);
static int parse_args_callback (char *, int, struct arg_options *);
static int output_line (Watch *, char *, int);
inline static int print_pending_and_line (Watch *, char *);
inline static int append_to_pending (Watch *, char *);

Watch *first_watch = NULL;
struct arg_options *global_opts;

static int
output_line (watch, str, len)
  Watch *watch;
  char *str;
  int len;
{
	struct tm linetime;

	if (len >= 16 && watch->flags & WATCH_FLAG_SYSLOG) {
		int c;
		syslog_time_to_tm(str, &linetime);
		for (c = 16; str[c] && !isspace(str[c]); c++); /* skip first word (computer name) */
		for (; str[c] && isspace(str[c]); c++); /* skip whitespace after that */
		if (!str[c]) return TRUE;
		str = &str[c];
		len -= c;
	} else {
		time_t secs;
		if (time(&secs) == -1) watch->flags &= ~WATCH_FLAG_TIMESTAMP;	/* stop timestamping of time fails */
		else memcpy(&linetime, localtime(&secs), sizeof(struct tm));
	}
	if (watch->flags & WATCH_FLAG_TIMESTAMP) {
		int tlen;
		char timestr[100];

		if (watch->timestamp)
			tlen = strftime(timestr, 100, watch->timestamp, &linetime);
		else
			tlen = strftime(timestr, 100, "%H:%M:%S ", &linetime);
		if (tlen > 0)
			if (!ui_print_line(watch, timestr, tlen, WATCH_FLAG_TIMESTAMP))
				return FALSE;
	}

	if (!ui_print_line(watch, str, len, 0))
		return FALSE;

	return TRUE;
}

/* update_watch:
 *   read and print data from a watch
 */
#define READ_BUFFER_SIZE 100
int
update_watch (watch)
  Watch *watch;
{
	char buf[READ_BUFFER_SIZE+1];
	char *inputbuf;
	ssize_t len;

	do {
		len = read(watch->fd, buf, READ_BUFFER_SIZE);
		if (len == -1) {
			if (errno == EAGAIN || errno == EWOULDBLOCK) break;
			inputbuf = NULL;
			len = asprintf(&inputbuf, "%s: Removing watch on %s (%s)\n", "monitor", watch->filename, strerror(errno));
			if (inputbuf) {
				if (!output_line(watch, inputbuf, len)) {
					free(inputbuf);
					return FALSE;
				}
				free(inputbuf);
			} else {
				printerror("Memory exhausted", NULL);
				return FALSE;
			}

			close_watch(watch);
		} else if (len > 0) {
			buf[len] = 0;
			inputbuf = buf;
			if (watch->flags) { /* must read whole line if parsing is enabled */
				char *line;
				while ((line = get_whole_line(&inputbuf))) {
					if (!print_pending_and_line(watch, line))
						return FALSE;
				}
				append_to_pending(watch, inputbuf);
			} else {
				if (!output_line(watch, buf, strlen(buf)))
					return FALSE;
			}
		}
	} while (len > 0);

	return TRUE;
}

Watch *
new_watch (filename, extra)
  char *filename;
  int extra;
{
	Watch *watch;

	/* allocate watch and link into list */
	watch = malloc(sizeof(Watch)+extra);
	if (!watch) {
		printerror("Memory exhausted", NULL);
		return NULL;
	}
	/* It is necessary to zero the memory (and -1 -> fd) if the allocation
	 * routines below fail. That way we can leave it in the list, and it
	 * will be automaticly deallocated when we exit. */
	memset(watch, 0, sizeof(Watch)+extra);
	watch->fd = -1;
	if (first_watch) watch->next = first_watch;
	else watch->next = NULL;
	first_watch = watch;

	/* set watch details */
	watch->delay = atoi(global_opts[OPT_DELAY].string);
	watch->flags = (global_opts[OPT_SYSLOG].exists ? WATCH_FLAG_SYSLOG : 0)
	  | (global_opts[OPT_TIMESTAMP].exists ? WATCH_FLAG_TIMESTAMP : 0);

	if (filename[0] == '-' && !filename[1]) {
		int flags;

		watch->filename = strdup("standard output");
		if (!watch->filename)
			return NULL;

		watch->fd = STDIN_FILENO;
		watch->type = WATCH_FIFO;

		/* configure STDIN to support nonblocking I/O operations */
		fcntl(STDIN_FILENO, F_GETFL, &flags);
		flags |= O_NONBLOCK;
		fcntl(STDIN_FILENO, F_SETFL, flags);
	} else {
		struct stat statbuf;

		watch->filename = strdup(filename);
		if (!watch->filename)
			return NULL;

		watch->fd = open(filename, O_RDONLY|O_NONBLOCK);
		if (watch->fd == -1) {
			printerror(filename, ": ", strerror(errno), NULL);
			return NULL;
		}

		if (fstat(watch->fd, &statbuf) == -1) {
			printerror(filename, ": ", strerror(errno), NULL);
			return NULL;
		}
		if (S_ISREG(statbuf.st_mode)) watch->type = WATCH_REGULAR;
		else if (S_ISFIFO(statbuf.st_mode)) watch->type = WATCH_FIFO;
		else if (S_ISCHR(statbuf.st_mode)) watch->type = WATCH_FIFO;
		else if (S_ISSOCK(statbuf.st_mode)) watch->type = WATCH_SOCKET;
		/* let the program be notified when I/O calls fail instead... */
		/*else {
			printerror(filename, ": cannot watch files of that type", NULL);
			return NULL;
		}*/
	}
	if (watch->type == WATCH_REGULAR) {
		if (lseek(watch->fd, 0, SEEK_END) == -1) {
			printerror(filename, ": ", strerror(errno), NULL);
			return NULL;
		}
		if (!monitor_timer(watch)) return NULL;
	} else {
		if (!monitor_file_event(watch)) return NULL;
	}

	/* copy timestamp string from option array */
	if (global_opts[OPT_TIMESTAMP].string && !(watch->timestamp = strdup(global_opts[OPT_TIMESTAMP].string)))
		return NULL;

	return watch;
}

/* parse_args: 
 *  returns exit code >=0 if we are to exit, or -1 if we are to continue
 */
inline static int
parse_args (argc, argv)
  int argc;
  char **argv;
{
	if (!arg_parser(argc-1, &argv[1], global_opts, parse_args_callback))
		return 1;

	/* quit with rc=0 if version is asked for */
	if (global_opts[OPT_VERSION].exists) {
		printf("%s (The Monitor) %s\n", ui_base_name, VERSION);
		return 0;
	}

	/* quit with rc=0 if help is asked for */
	if (global_opts[OPT_HELP].exists) {
		int maxlen=0,prtlen,c;
		printf("Usage: %s [OPTION]... FILE...\n\
Displays the output of one or several files (regular, fifos, sockets etc),\n\
or special devices (clock).\n\n", argv[0]);
		/* count max length for the option-column, in characters */
		for(c = 0; global_opts[c].long_name; c++) {
			maxlen = MAX(maxlen,
			  strlen(global_opts[c].long_name)
			  + (global_opts[c].arg_help ? strlen(global_opts[c].arg_help) : 0));
		}
		for(c = 0; global_opts[c].long_name; c++) {
			if (global_opts[c].abbreviation) printf("  -%c,", global_opts[c].abbreviation);
			else printf("     ");
			prtlen = strlen(global_opts[c].long_name);
			if (global_opts[c].arg_help) {
				printf(" --%s%s", global_opts[c].long_name, global_opts[c].arg_help);
				prtlen += strlen(global_opts[c].arg_help);
			}
			else printf(" --%s", global_opts[c].long_name);
			for(; prtlen < maxlen+2; prtlen++) printf(" ");
			puts(global_opts[c].help);
		}
		puts("\nBy default, if no parameter for timestamp is specified, \"%H:%M:%S \" is assumed.\n\
Default delay for regular files is 2000 milliseconds.\n\
The order of some arguments is important - they affect the following watch\n\
(file) and timestamp arguments. --syslog implies removal (and parsing) of\n\
date, host.\n\nReport bugs to osk@hem.passagen.se");
		return 0;
	}

	return -1;
}

static int
parse_args_callback(str, opt, opts)
	char *str;
	int opt;
	struct arg_options *opts;
{

	if (opts) {
		/* don't process options if we already have help or version options */
		if (opts[OPT_HELP].exists || opts[OPT_VERSION].exists)
			return TRUE;
		else {
			if (opt == OPT_NO_SYSLOG
			  || opt == OPT_NO_TIMESTAMP) {
				opts[opt-1].string = NULL;
				opts[opt-1].exists = FALSE;
			}
			return ui_parse_args_callback(str, opt, opts);
		}
	} else {
		fprintf(stderr, "%s: %s\n", ui_base_name, str);
	}

	return TRUE;
}

/* recursively close all watches */
void
close_watches (void)
{
	Watch *watch;
	Watch *nextwatch;

	watch = first_watch;
	while (watch) {
		nextwatch = watch->next;
		close_watch(watch);
		watch = nextwatch;
	}
}

void
close_watch (watch)
  Watch *watch;
{
	Watch *tmpwatch;

	ui_close_watch(watch);

	/* unlink from list */
	tmpwatch = first_watch;
	if (first_watch == watch) first_watch = watch->next;
	else {
		while (tmpwatch) {
			if (tmpwatch->next == watch) { tmpwatch->next = watch->next; break; }
			tmpwatch = tmpwatch->next;
		}
	}

	if (watch->fd != -1) {
		if (watch->type & WATCH_TYPE_EVENT) unmonitor_file_event(watch);
		else unmonitor_timer(watch);
		close(watch->fd);
	}
	if (watch->pending) free(watch->pending);
	if (watch->timestamp) free(watch->timestamp);
	if (watch->filename) free(watch->filename);
	free(watch);
}

RETSIGTYPE
terminate_signal(sig)
  int sig;
{
	close_watches();
	exit_ui();
}

/* print_pending_and_line:
 *   prints pending (if any) and first line in argument "line".
 */
inline static int
print_pending_and_line(watch, line)
  Watch *watch;
  char *line;
{
	int len;

	/* we can assume that LF comes before null here, because we called
	 * get_whole_line first
	 */
	for (len=0; line[len] != '\n'; len++);
	len++;

	if (watch->pending) {
		watch->pending = realloc(watch->pending, watch->pending_len+len);
		if (!watch->pending) {
			printerror("Memory exhausted", NULL);
			return FALSE;
		}
		memcpy(&(watch->pending)[watch->pending_len], line, len);
		if (!output_line(watch, watch->pending, len+watch->pending_len))
			return FALSE;
		free(watch->pending);
		watch->pending = NULL;
		watch->pending_len = 0;
	} else {
		if (!output_line(watch, line, len))
			return FALSE;
	}
	
	return TRUE;
}

/* append_to_pending:
 *   append data to pending line.
 */
inline static int
append_to_pending(watch, line)
  Watch *watch;
  char *line;
{
	if (line[0]) {
		int len = strlen(line);

		if (watch->pending) {
			watch->pending = realloc(watch->pending, watch->pending_len+len+1);
			if (!watch->pending) {
				printerror("Memory exhausted", NULL);
				return FALSE;
			}
			memcpy(&(watch->pending)[watch->pending_len], line, len);
			watch->pending_len += len;
		} else {
			watch->pending = malloc(len);
			if (!watch->pending) {
				printerror("Memory exhausted", NULL);
				return FALSE;
			}
			memcpy(watch->pending, line, len);
			watch->pending_len = len;
		}
	} else {
		watch->pending = NULL;
	}

	return TRUE;
}

int
main (argc, argv)
  int argc;
  char **argv;
{
	static struct arg_options monitor_opts[] = {
	/* per-watch options */
		{ "timestamp", 't', 	ARG_OPTIONAL, FALSE, NULL,		"[=FORMAT]", "timestamp lines read (see strftime(3) for format)" },
		{ "no-timestamp", 0, 	ARG_NONE, 	  FALSE, NULL,		NULL, "do not print timestamp" },
		{ "syslog", 's', 		ARG_NONE, 	  FALSE, NULL, 		NULL, "enable syslog parsing" },
		{ "no-syslog", 0, 		ARG_NONE, 	  FALSE, NULL, 		NULL, "disable syslog parsing" },
		{ "delay", 0,			ARG_REQUIRED, TRUE,  "2000", 	"=MILLISECONDS", "delay between reads (in milliseconds)" },
	/* misc other options */
//		{ "config", 0,			ARG_REQUIRED, FALSE, NULL, 		"=FILE", "read options from specified file" },
		{ "help", 0, 			ARG_NONE, 	  FALSE, NULL, 		NULL, "display this help and exit" },
		{ "version", 'v', 		ARG_NONE, 	  FALSE, NULL, 		NULL, "output version information and exit" },
		{ NULL, 0, 0, 0, NULL, NULL, NULL }
	};
	struct sigaction action;
	sigset_t set;
	int rc;

	/* initialize gui */
	if (!initialize_ui(&argc, &argv))
		return 1;

	/* install signal handlers */
	action.sa_handler = SIG_IGN;
	sigemptyset(&set);
	action.sa_mask = set;
	action.sa_flags = SA_RESTART;
	if (sigaction(SIGPIPE, &action, NULL)
	    || sigaction(SIGHUP, &action, NULL)) {
		printerror(strerror(errno), NULL);
		exit_ui();
		return 1;
	}

	action.sa_handler = terminate_signal;
	sigaddset(&set, SIGTERM);
	sigaddset(&set, SIGUSR1);
	if (sigaction(SIGTERM, &action, NULL)) {
		printerror(strerror(errno), NULL);
		exit_ui();
		return 1;
	}

	/* merge option arrays */
	if (ui_opts[0].long_name) {
		int c,c2;

		/* count total number of options */
		for (c = 0; monitor_opts[c].long_name; c++);
		for (c2 = 0; ui_opts[c2].long_name; c2++);
		c += c2 + 1;

		global_opts = malloc(sizeof(struct arg_options) * c);
		if (!global_opts) {
			printerror("Memory exhausted", NULL);
			exit_ui();
			return 1;
		}

		for(c = 0; monitor_opts[c].long_name; c++)
			global_opts[c] = monitor_opts[c];
		for(c2 = 0; ui_opts[c2].long_name; c2++) {
			global_opts[c] = ui_opts[c2];
			c++;
		}
		global_opts[c].long_name = NULL;
	} else {
		global_opts = monitor_opts;
	}

	/* parse command line arguments */
	if ((rc = parse_args(argc, argv)) != -1) {
		exit_ui();
		close_watches();
		return rc;
	}
	
	/* start the application and the gui */
	if (!enter_ui_mainloop(global_opts)) {
		exit_ui();
		close_watches();
		return 1;
	}

	exit_ui();
	close_watches();

	/* There is no point in free(global_opts).
	 * global_opts is used and required throughout the application,
	 * and will be freed automaticly once we exit. */

	return 0;
}
